import json
import time
import unittest

from hamcrest import assert_that
from hamcrest import is_

from netman.adapters.switches.remote import RemoteSwitch
from netman.adapters.threading_lock_factory import ThreadingLockFactory
from netman.core.objects.flow_control_switch import FlowControlSwitch
from tests.system_tests import NetmanTestApp, get_available_switch, create_session


class SessionTest(unittest.TestCase):

    def test_creating_commit_deleting_session_works(self):
        with NetmanTestApp() as partial_client:
            client = partial_client(get_available_switch("cisco"))

            result = create_session(client, "i_love_sessions")
            session_id = result.json()['session_id']
            client.post("/switches-sessions/" + session_id + "/actions", data='start_transaction')

            result = client.post("/switches-sessions/" + session_id + "/actions", data='commit')
            assert_that(result.status_code, is_(204), result.text)

            client.post("/switches-sessions/" + session_id + "/actions", data='end_transaction')
            result = client.delete("/switches-sessions/" + session_id)
            assert_that(result.status_code, is_(204), result.text)

    def test_sessions_can_time_out(self):
        with NetmanTestApp() as partial_client:
            client = partial_client(get_available_switch("brocade"))

            result = create_session(client, "i_love_sessions")

            session_id = result.json()['session_id']

            time.sleep(2)

            result = client.post("/switches-sessions/" + session_id + "/actions", data='start_transaction')
            assert_that(result.json().get("error"), is_('Session "i_love_sessions" not found.'))
            assert_that(result.status_code, is_(404), 'Session should have timed out')

    def test_creating_two_duplicate_sessions_returns_409(self):
        with NetmanTestApp() as partial_client:
            client = partial_client(get_available_switch("dell"))

            result = create_session(client, "i_love_sessions")

            first_session_id = result.json()['session_id']

            result = client.post("/switches-sessions/{}".format(first_session_id),
                                 data=json.dumps({"hostname": client.switch.hostname}))
            assert_that(result.status_code, is_(409), result.text)

    def test_remote_sessions_can_continue_on_a_different_netman(self):
        with NetmanTestApp() as partial_client1, NetmanTestApp() as partial_client2:
            switch_descriptor = get_available_switch("juniper")

            client1 = partial_client1(switch_descriptor)
            first_netman_url = "{}:{}".format(client1.host, client1.port)

            client2 = partial_client2(switch_descriptor)
            second_netman_url = "{}:{}".format(client2.host, client2.port)

            remote_switch = RemoteSwitch(switch_descriptor)
            remote_switch._proxy = first_netman_url
            switch = FlowControlSwitch(remote_switch, ThreadingLockFactory().new_lock())

            with switch.transaction():
                switch.add_vlan(1498, "one")

                remote_switch._proxy = second_netman_url

                switch.add_vlan(1499, "two")

            assert_that(client1.get("/switches/{hostname}/vlans/1498").json()["name"], is_("one"))
            assert_that(client1.get("/switches/{hostname}/vlans/1499").json()["name"], is_("two"))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fake_switches.arista.arista_core import AristaSwitchCore
from fake_switches.brocade.brocade_core import BrocadeSwitchCore
from fake_switches.cisco.cisco_core import CiscoSwitchCore
from fake_switches.cisco6500.cisco_core import Cisco6500SwitchCore
from fake_switches.dell.dell_core import DellSwitchCore
from fake_switches.dell10g.dell_core import Dell10GSwitchCore
from fake_switches.juniper.juniper_core import JuniperSwitchCore
from fake_switches.juniper_mx.juniper_mx_core import JuniperMXSwitchCore
from fake_switches.juniper_qfx_copper.juniper_qfx_copper_core import JuniperQfxCopperSwitchCore
from fake_switches.transports.ssh_service import SwitchSshService
from fake_switches.switch_configuration import Port, AggregatedPort
from fake_switches.transports.telnet_service import SwitchTelnetService
from fake_switches.transports.http_service import SwitchHttpService

from netman.core.objects.switch_descriptor import SwitchDescriptor

available_models = [
    {
        "switch_descriptor": SwitchDescriptor(
            model="arista_http",
            hostname="127.0.0.1",
            port=11015,
            username="root",
            password="root",
        ),
        "test_port_name": "Ethernet1",
        "test_vrrp_track_id": "",
        "core_class": AristaSwitchCore,
        "service_class": SwitchHttpService,
        "ports": [
            Port("Ethernet1"),
            Port("Ethernet2"),
        ]
    },
    {
        "switch_descriptor": SwitchDescriptor(
            model="cisco",
            hostname="127.0.0.1",
            port=11002,
            username="root",
            password="root",
        ),
        "test_port_name": "FastEthernet0/3",
        "test_vrrp_track_id": "101",
        "core_class": CiscoSwitchCore,
        "service_class": SwitchSshService,
        "ports": [
            Port("FastEthernet0/1"),
            Port("FastEthernet0/2"),
            Port("FastEthernet0/3"),
            Port("FastEthernet0/4"),
        ]
    },
    {
        "switch_descriptor": SwitchDescriptor(
            model="cisco",
            hostname="127.0.0.1",
            port=11014,
            username="root",
            password="root",
        ),
        "test_port_name": "FastEthernet0/4",
        "test_vrrp_track_id": "102",
        "core_class": Cisco6500SwitchCore,
        "service_class": SwitchSshService,
        "ports": [
            Port("FastEthernet0/1"),
            Port("FastEthernet0/2"),
            Port("FastEthernet0/3"),
            Port("FastEthernet0/4"),
        ]
    },
    {
        "switch_descriptor": SwitchDescriptor(
            model="brocade",
            hostname="127.0.0.1",
            port=11003,
            username="root",
            password="root",
        ),
        "test_port_name": "ethernet 1/3",
        "test_vrrp_track_id": "ethernet 1/1",
        "core_class": BrocadeSwitchCore,
        "service_class": SwitchSshService,
        "ports": [
            Port("ethernet 1/1"),
            Port("ethernet 1/2"),
            Port("ethernet 1/3"),
            Port("ethernet 1/4")
        ]
    },
    {
        "switch_descriptor": SwitchDescriptor(
            model="brocade_telnet",
            hostname="127.0.0.1",
            port=11012,
            username="root",
            password="root",
        ),
        "test_port_name": "ethernet 1/3",
        "test_vrrp_track_id": "ethernet 1/1",
        "core_class": BrocadeSwitchCore,
        "service_class": SwitchTelnetService,
        "ports": [
            Port("ethernet 1/1"),
            Port("ethernet 1/2"),
            Port("ethernet 1/3"),
            Port("ethernet 1/4")
        ]
    },
    {
        "switch_descriptor": SwitchDescriptor(
            model="juniper",
            hostname="127.0.0.1",
            port=11004,
            username="root",
            password="root",
        ),
        "test_port_name": "ge-0/0/3",
        "core_class": JuniperSwitchCore,
        "service_class": SwitchSshService,
        "ports": [
            Port("ge-0/0/1"),
            Port("ge-0/0/2"),
            Port("ge-0/0/3"),
            Port("ge-0/0/4"),
            AggregatedPort("ae1"),
            AggregatedPort("ae2"),
            AggregatedPort("ae3"),
            AggregatedPort("ae4"),
            AggregatedPort("ae42")
        ]
    },
    {
        "switch_descriptor": SwitchDescriptor(
            model="juniper_qfx_copper",
            hostname="127.0.0.1",
            port=11005,
            username="root",
            password="root",
        ),
        "test_port_name": "ge-0/0/3",
        "core_class": JuniperQfxCopperSwitchCore,
        "service_class": SwitchSshService,
        "ports": [
            Port("ge-0/0/1"),
            Port("ge-0/0/2"),
            Port("ge-0/0/3"),
            Port("ge-0/0/4"),
            AggregatedPort("ae1"),
            AggregatedPort("ae2"),
            AggregatedPort("ae3"),
            AggregatedPort("ae4"),
            AggregatedPort("ae42")
        ]
    },
    {
        "switch_descriptor": SwitchDescriptor(
            model="dell",
            hostname="127.0.0.1",
            port=11006,
            username="root",
            password="root",
        ),
        "test_port_name": "ethernet 2/g2",
        "core_class": DellSwitchCore,
        "service_class": SwitchSshService,
        "ports": [
            Port("ethernet 1/g1"),
            Port("ethernet 1/g2"),
            Port("ethernet 1/xg1"),
            Port("ethernet 2/g1"),
            Port("ethernet 2/g2"),
            Port("ethernet 2/xg1")
        ]
    },
    {
        "switch_descriptor": SwitchDescriptor(
            model="dell_telnet",
            hostname="127.0.0.1",
            port=11007,
            username="root",
            password="root",
        ),
        "test_port_name": "ethernet 2/g2",
        "core_class": DellSwitchCore,
        "service_class": SwitchTelnetService,
        "ports": [
            Port("ethernet 1/g1"),
            Port("ethernet 1/g2"),
            Port("ethernet 1/xg1"),
            Port("ethernet 2/g1"),
            Port("ethernet 2/g2"),
            Port("ethernet 2/xg1")
        ]
    },
    {
        "switch_descriptor": SwitchDescriptor(
            model="dell10g",
            hostname="127.0.0.1",
            port=11008,
            username="root",
            password="root",
        ),
        "test_port_name": "tengigabitethernet 1/0/1",
        "core_class": Dell10GSwitchCore,
        "service_class": SwitchSshService,
        "ports": [
            Port("tengigabitethernet 1/0/1"),
            Port("tengigabitethernet 1/0/2"),
            Port("tengigabitethernet 2/0/1"),
            Port("tengigabitethernet 2/0/2"),
        ]
    },
    {
        "switch_descriptor": SwitchDescriptor(
            model="dell10g_telnet",
            hostname="127.0.0.1",
            port=11009,
            username="root",
            password="root",
        ),
        "test_port_name": "tengigabitethernet 1/0/1",
        "core_class": Dell10GSwitchCore,
        "service_class": SwitchTelnetService,
        "ports": [
            Port("tengigabitethernet 1/0/1"),
            Port("tengigabitethernet 1/0/2"),
            Port("tengigabitethernet 2/0/1"),
            Port("tengigabitethernet 2/0/2"),
        ]
    },
    {
        "switch_descriptor": SwitchDescriptor(
            model="juniper_mx",
            hostname="127.0.0.1",
            port=11010,
            username="root",
            password="root",
        ),
        "test_port_name": "xe-0/0/3",
        "core_class": JuniperMXSwitchCore,
        "service_class": SwitchSshService,
        "ports": [
            Port("xe-0/0/1"),
            Port("xe-0/0/2"),
            Port("xe-0/0/3"),
            Port("xe-0/0/4"),
            AggregatedPort("ae1"),
            AggregatedPort("ae2"),
            AggregatedPort("ae3"),
            AggregatedPort("ae4"),
            AggregatedPort("ae42")
        ]
    }
]



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ncclient.xml_ import to_ele

from netman.adapters.switches.juniper.base import interface_replace, bond_name, Juniper
from netman.adapters.switches.juniper.standard import JuniperCustomStrategies


def netconf(switch_descriptor):
    return Juniper(switch_descriptor, custom_strategies=JuniperQfxCopperCustomStrategies())


class JuniperQfxCopperCustomStrategies(JuniperCustomStrategies):
    def get_interface_port_mode_update_element(self, mode):
        return to_ele("<interface-mode>{}</interface-mode>".format(mode))

    def get_port_mode_node_in_inteface_node(self, interface_node):
        return interface_node.xpath("unit/family/ethernet-switching/interface-mode")

    def add_enslave_to_bond_operations(self, update, interface, bond):
        ether_options = [
            to_ele("<auto-negotiation/>"),
            to_ele("""
                <ieee-802.3ad>
                    <bundle>{0}</bundle>
                </ieee-802.3ad>
            """.format(bond_name(bond.number)))]

        update.add_interface(interface_replace(interface, *ether_options))

    def add_update_bond_members_speed_operations(self, update, slave_nodes, speed):
        pass

    def get_interface_trunk_native_vlan_id_node(self, interface):
        return interface.xpath("native-vlan-id")

    def set_native_vlan_id_node(self, interface_node, native_vlan_id_node):
        return interface_node.xpath("//interface")[0].append(native_vlan_id_node)

    def get_protocols_interface_name(self, interface_name):
        return interface_name



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netman.core.objects import Model


class BaseInterface(Model):
    def __init__(self, shutdown=None, port_mode=None, access_vlan=None,
                 trunk_native_vlan=None, trunk_vlans=None, mtu=None):
        self.shutdown = shutdown
        self.port_mode = port_mode
        self.access_vlan = access_vlan
        self.trunk_native_vlan = trunk_native_vlan
        self.trunk_vlans = trunk_vlans or []
        self.mtu = mtu


class Interface(BaseInterface):
    def __init__(self, name=None, bond_master=None, auto_negotiation=None, **interface):
        super(Interface, self).__init__(**interface)
        self.name = name
        self.bond_master = bond_master
        self.auto_negotiation = auto_negotiation



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import warnings
from functools import wraps

from hamcrest.core.core.isequal import IsEqual
from netaddr import IPNetwork

from global_reactor import ThreadedReactor
from tests.adapters.model_list import available_models
from tests.adapters.shell.telnet_login_special_cases_test import hanging_password_telnet_hook
from tests.adapters.shell.terminal_client_test import telnet_hook_to_reactor, ssh_hook_to_reactor


def setup():
    ThreadedReactor.start_reactor(available_models, reactor_hook_callbacks=[
        hanging_password_telnet_hook,
        telnet_hook_to_reactor,
        ssh_hook_to_reactor
    ])


def tearDown():
    ThreadedReactor.stop_reactor()


class ExactIpNetwork(object):
    def __init__(self, ip, mask=None):
        net = IPNetwork(ip)
        self.ip = net.ip
        self.mask = mask or net.prefixlen

    def __eq__(self, other):
        if not isinstance(other, IPNetwork):
            logging.error("{} is not an IPNetwork object".format(repr(other)))
            return False
        result = other.ip == self.ip and other.prefixlen == self.mask
        if not result:
            logging.error("Expected {}/{} and got {}/{}".format(self.ip, self.mask, other.ip, other.prefixlen))
        return result

    def __repr__(self):
        return "{}/{}".format(self.ip, self.mask)


def ignore_deprecation_warnings(func):
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', DeprecationWarning)
            ret = func(*args, **kwargs)
            return ret
    return func_wrapper


def has_message(msg):
    return HasMessage(msg)


class HasMessage(IsEqual):
    def _matches(self, item):
        return super(HasMessage, self)._matches(str(item))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest
import flask
from netman import raw_or_json

from tests.api import open_fixture


class BaseApiTest(unittest.TestCase):
    def setUp(self):
        super(BaseApiTest, self).setUp()

        self.app = flask.Flask(__name__)

    def get(self, url, **kwargs):
        with self.app.test_client() as http_client:
            request_result = http_client.get(url, **kwargs)

        return to_json(request_result.data), request_result.status_code

    def post(self, url, data=None, raw_data=None, fixture=None, **kwargs):
        if fixture is not None:
            posting_data = open_fixture(fixture).read()
        else:
            posting_data = raw_or_json(raw_data, data)

        with self.app.test_client() as http_client:
            request_result = http_client.post(url, data=posting_data, **kwargs)
        return to_json(request_result.data), request_result.status_code

    def delete(self, url, **kwargs):
        with self.app.test_client() as http_client:
            request_result = http_client.delete(url, **kwargs)

        return to_json(request_result.data), request_result.status_code

    def put(self, url, data=None, raw_data=None, fixture=None, **kwargs):
        if fixture is not None:
            posting_data = open_fixture(fixture).read()
        else:
            posting_data = raw_or_json(raw_data, data)

        with self.app.test_client() as http_client:
            request_result = http_client.put(url, data=posting_data, **kwargs)

        return to_json(request_result.data), request_result.status_code


def to_json(string):
    try:
        return json.loads(string)
    except ValueError:
        return None



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

import unittest

from flexmock import flexmock
from hamcrest import is_, equal_to
from hamcrest import assert_that

from netman.api.api_utils import MultiContext, BadRequest
from netman.api.validators import is_vlan_number, is_boolean, Vlan, Interface, is_dict_with, optional, is_type
from netman.core.objects.exceptions import BadVlanNumber


class DecoratorTests(unittest.TestCase):
    def test_resource_decorator(self):
        process1 = flexmock()
        process2 = flexmock()

        process1.should_receive('process').with_args('arg2').once()
        process1.should_receive('__enter__').once()

        process2.should_receive('process').with_args('arg2').once()
        process2.should_receive('__enter__').once()

        process1.should_receive('__exit__').once()
        process2.should_receive('__exit__').once()

        with MultiContext('arg1', 'arg2', process1, process2):
            pass

    def test_content_vlan(self):
        self.assertEquals(is_vlan_number('28'), {'vlan_number': 28})

    def test_content_vlan_invalid(self):
        self.assertRaises(BadVlanNumber, is_vlan_number, 'patate')

    def test_content_vlan_invalid_number(self):
        self.assertRaises(BadVlanNumber, is_vlan_number, 2888888)

    def test_content_shutdown_options_true(self):
        self.assertEquals(is_boolean('true'), {'state': True})

    def test_content_shutdown_options_false(self):
        self.assertEquals(is_boolean('false'), {'state': False})

    def test_content_shutdown_options_invalid(self):
        self.assertRaises(BadRequest, is_boolean, 'patate')

    def test_resource_vlan(self):
        vlan = Vlan(None)
        vlan.process({'vlan_number': 2999})
        self.assertEquals(vlan.__enter__(), 2999)

    def test_resource_interface(self):
        interface = Interface(None)
        interface.process({'interface_id': 28})
        self.assertEquals(interface.__enter__(), 28)


class IsDictTests(unittest.TestCase):
    def test_empty(self):
        validate = is_dict_with()

        assert_that(validate("{}"), is_({}))

    def test_one_boolean(self):
        validate = is_dict_with(mybool=is_type(bool))
        data = {
            "mybool": True
        }

        assert_that(validate(json.dumps(data)), is_(data))

    def test_optional(self):
        validate = is_dict_with(mybool=optional(is_type(bool)))

        assert_that(validate("{}"), is_({}))

    def test_extra_field(self):
        validate = is_dict_with(mybool=is_type(bool))
        data = {
            "mybool": True,
            "yourbool": True
        }

        with self.assertRaises(BadRequest) as expect:
            validate(json.dumps(data))

        assert_that(str(expect.exception), equal_to("Unknown key: yourbool"))

    def test_incorrect_field(self):
        validate = is_dict_with(mybool=is_type(bool))
        data = {
            "mybool": "string",
        }

        with self.assertRaises(BadRequest) as expect:
            validate(json.dumps(data))

        assert_that(str(expect.exception), equal_to("Expected \"bool\" type for key mybool, got \"unicode\""))

    def test_not_json(self):
        validate = is_dict_with()

        with self.assertRaises(BadRequest) as expect:
            validate("what")

        assert_that(str(expect.exception), equal_to("Malformed JSON request"))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings
from functools import wraps

from netman.core.objects.flow_control_switch import FlowControlSwitch


def transactional(fn):
    warnings.warn("Deprecated, make your own annotation this one will disappear", DeprecationWarning)

    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        if self.in_transaction:
            return fn(self, *args, **kwargs)
        else:
            with self.transaction():
                return fn(self, *args, **kwargs)

    return wrapper


class SwitchTransactional(FlowControlSwitch):
    def __init__(self, impl, lock):
        warnings.warn("Deprecated, use FlowControlSwitch instead", DeprecationWarning)
        super(SwitchTransactional, self).__init__(impl, lock)

    @property
    def in_transaction(self):
        return self.wrapped_switch.in_transaction



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from _socket import timeout, gaierror
import re
import telnetlib
from telnetlib import IAC, DO, DONT, WILL, WONT

from netman.adapters import shell
from netman.adapters.shell.base import TerminalClient
from netman.core.objects.exceptions import CouldNotConnect, CommandTimeout, ConnectTimeout


class TelnetClient(TerminalClient):

    def __init__(self, host, username, password, port=23, prompt=('>', '#'),
                 connect_timeout=None, command_timeout=None, **_):
        self.prompt = prompt
        self.host = host
        self.port = port
        self.command_timeout = command_timeout or shell.default_command_timeout
        self.connect_timeout = connect_timeout or shell.default_connect_timeout
        self.full_log = ""

        self.telnet = self._connect()
        self._login(username, password)

    def do(self, command, wait_for=None, include_last_line=False):
        self.telnet.write(str(command) + "\r\n")
        result = self._read_until(wait_for)

        return _filter_input_and_empty_lines(command, include_last_line, result)

    def send_key(self, key, wait_for=None, include_last_line=False):
        self.telnet.write(key)
        result = self._read_until(wait_for)

        return _filter_input_and_empty_lines(key, include_last_line, result)

    def quit(self, command):
        self.telnet.write(command + "\r\n")

    def get_current_prompt(self):
        return self.full_log.splitlines()[-1]

    def _login(self, username, password):
        self.telnet.read_until(":", self.command_timeout)
        self.telnet.write(str(username) + "\r\n")
        self.telnet.read_until(":", self.command_timeout)
        self.telnet.write(str(password) + "\r\n")

        result = self._wait_for_successful_login()
        self.full_log += result[len(password):].lstrip()

    def _read_until(self, wait_for):
        expect = wait_for or self.prompt
        if isinstance(expect, basestring):
            expect = [expect]
        expect = ["{}$".format(re.escape(s)) for s in list(expect)]

        result = self._wait_for(expect)
        self.full_log += result

        return result

    def _wait_for(self, expect):
        result = self.telnet.expect(expect, timeout=self.command_timeout)
        if result[0] == -1:
            raise CommandTimeout(expect)
        return result[2]

    def _wait_for_successful_login(self):
        result = self.telnet.expect(list(self.prompt), timeout=self.connect_timeout)
        if result[0] == -1:
            raise ConnectTimeout(self.host, self.port)
        return result[2]

    def _connect(self):
        try:
            telnet = telnetlib.Telnet(self.host, self.port, self.connect_timeout)
        except timeout:
            raise ConnectTimeout(self.host, self.port)
        except gaierror:
            raise CouldNotConnect(self.host, self.port)

        telnet.set_option_negotiation_callback(_accept_all)
        return telnet


def _accept_all(sock, cmd, opt):
    if cmd == WILL:
        sock.sendall(IAC + DO + opt)
    elif cmd == WONT:
        sock.sendall(IAC + DONT + opt)


def _filter_input_and_empty_lines(command, include_last_line, result):
    lines = result[len(command):].splitlines()
    if not include_last_line:
        lines = lines[:-1]
    return filter(None, lines)



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def sub_dict(d, *keys):
    return {k: d[k] for k in keys}


class Serializer(object):
    since_version = 1

    def to_api(self, core_object):
        raise NotImplementedError()

    def to_core(self, api_dict):
        raise NotImplementedError()


class Serializers(object):
    def __init__(self, *serializers):
        self.serializers = sorted(serializers, key=lambda s: s.since_version, reverse=True)

    def at_most(self, version):
        if version is None:
            return self.serializers[-1]
        for s in self.serializers:
            if s.since_version <= float(version):
                return s

    def to_api(self, core_object, version=None):
        return self.at_most(version).to_api(core_object)

    def to_core(self, api_dict, version=None):
        return self.at_most(version).to_core(api_dict)



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from hamcrest import assert_that, is_
from netman.adapters import shell
from twisted.internet.protocol import Factory

from netman.adapters.shell.telnet import TelnetClient
from netman.core.objects.exceptions import ConnectTimeout
from tests.adapters.shell.mock_telnet import MockTelnet


class TelnetLoginSpecialCasesTest(unittest.TestCase):
    client = TelnetClient
    port = 10013

    def test_hanging_during_the_login_process_raises_a_connect_timeout(self):
        shell.default_connect_timeout = 0.1
        with self.assertRaises(ConnectTimeout) as expect:
            self.client("127.0.0.1", "admin", "1234", self.port)

        assert_that(str(expect.exception), is_("Timed out while connecting to 127.0.0.1 on port 10013"))


class PasswordHangingMockTelnet(MockTelnet):
    def validate_password(self, _):
        pass


class SwitchTelnetFactory(Factory):
    def __init__(self, prompt, commands):
        self.prompt = prompt
        self.commands = commands

    def protocol(self):
        return PasswordHangingMockTelnet(prompt=self.prompt, commands=self.commands)


def hanging_password_telnet_hook(reactor):
    reactor.listenTCP(interface="127.0.0.1",
                      port=TelnetLoginSpecialCasesTest.port,
                      factory=SwitchTelnetFactory("hostname>", []))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MemoryStorage(object):
    def __init__(self):
        self.data = {}

    def set(self, name, data):
        self.data[name] = data

    def get(self, name):
        return self.data[name]

    def add_switch_descriptor(self, switch_descriptor):
        self.set("SWITCH:" + switch_descriptor.hostname, switch_descriptor)

    def get_switch_descriptor(self, hostname):
        return self.get("SWITCH:" + hostname)

    def get_switches(self):
        return [v for k, v in self.data.items() if k.startswith("SWITCH:")]



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tests.adapters.model_list import available_models
from tests.adapters.unified_tests.bond_management_test import BondManagementTest
from tests.adapters.unified_tests.interface_management_test import InterfaceManagementTest
from tests.adapters.unified_tests.ip_management_test import IpManagementTest
from tests.adapters.unified_tests.vlan_management_test import VlanManagementTest
from tests.adapters.unified_tests.vrrp_test import VrrpTest
from tests.adapters.unified_tests.dhcp_relay_server_test import DhcpRelayServerTest


test_classes = [
    BondManagementTest,
    InterfaceManagementTest,
    IpManagementTest,
    VlanManagementTest,
    VrrpTest,
    DhcpRelayServerTest
]


current_module = globals()
for specs in available_models:
    for test_class in test_classes:
        class_name = "{}{}".format(specs["switch_descriptor"].model.capitalize(), test_class.__name__)
        current_module[class_name] = type(class_name, (test_class,), {"__test__": True, "switch_specs": specs})



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import TestCase

from flexmock import flexmock
from hamcrest import assert_that, is_

from netman.core.objects.exceptions import NetmanException
from netman.core.objects.switch_base import SwitchBase
from netman.core.objects.switch_descriptor import SwitchDescriptor
from netman.core.objects import switch_transactional
from tests import ignore_deprecation_warnings


class SwitchTransactionalTest(TestCase):
    """
    Backward compatibility test
    """

    @ignore_deprecation_warnings
    def setUp(self):
        self.switch_impl = flexmock(SwitchBase(SwitchDescriptor("cisco", "name")))
        self.switch_impl.connected = True
        self.lock = flexmock()
        self.switch = switch_transactional.SwitchTransactional(self.switch_impl, self.lock)

    @ignore_deprecation_warnings
    def test_transactional_annotation_does_nothing_in_a_transaction(self):

        self.lock.should_receive("acquire").with_args().once().ordered()
        self.switch_impl.should_receive("_start_transaction").with_args().once().ordered()

        with self.switch.transaction():
            assert_that(self.switch.in_transaction, is_(True))

            result = self.decorated_method(self.switch)
            assert_that(result, is_("Good"))

            result = self.decorated_method(self.switch)
            assert_that(result, is_("Good"))

            self.switch_impl.should_receive("commit_transaction").with_args().once().ordered()
            self.switch_impl.should_receive("_end_transaction").with_args().once().ordered()
            self.lock.should_receive("release").with_args().once().ordered()

        assert_that(self.switch.in_transaction, is_(False))

    def test_lock_is_released_if_start_transaction_fails(self):
        self.lock.should_receive("acquire").with_args().once().ordered()
        self.switch_impl.should_receive("_start_transaction").with_args().and_raise(NetmanException()).once().ordered()
        self.lock.should_receive("release").with_args().once().ordered()

        with self.assertRaises(NetmanException):
            with self.switch.transaction():
                pass

        assert_that(self.switch.in_transaction, is_(False))

    @ignore_deprecation_warnings
    def test_transactional_annotation_runs_the_method_in_a_transaction_when_not_in_one(self):

        self.lock.should_receive("acquire").with_args().once().ordered()
        self.switch_impl.should_receive("_start_transaction").with_args().once().ordered()
        self.switch_impl.should_receive("commit_transaction").with_args().once().ordered()
        self.switch_impl.should_receive("_end_transaction").with_args().once().ordered()
        self.lock.should_receive("release").with_args().once().ordered()

        result = self.decorated_method(self.switch)

        assert_that(self.switch.in_transaction, is_(False))

        assert_that(result, is_("Good"))

    @ignore_deprecation_warnings
    def test_lock_is_released_if_start_transaction_fails_using_annotation(self):
        self.lock.should_receive("acquire").with_args().once().ordered()
        self.switch_impl.should_receive("_start_transaction").with_args().and_raise(NetmanException()).once().ordered()
        self.lock.should_receive("release").with_args().once().ordered()

        with self.assertRaises(NetmanException):
            self.decorated_method(self.switch)

        assert_that(self.switch.in_transaction, is_(False))

    @property
    def decorated_method(self):
        @switch_transactional.transactional
        def transactional_method(self):
            if self.in_transaction is False:
                raise AssertionError("I'm not in a transaction")
            return "Good"
        return transactional_method



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import warnings

from netaddr.ip import IPNetwork, IPAddress

from netman import regex
from netman.adapters.shell.ssh import SshClient
from netman.adapters.switches.util import SubShell, split_on_dedent, split_on_bang, no_output
from netman.core.objects.access_groups import IN, OUT
from netman.core.objects.exceptions import IPNotAvailable, UnknownVlan, UnknownIP, UnknownAccessGroup, BadVlanNumber, \
    BadVlanName, UnknownInterface, UnknownVrf, VlanVrfNotSet, IPAlreadySet, VrrpAlreadyExistsForVlan, BadVrrpGroupNumber, \
    BadVrrpPriorityNumber, VrrpDoesNotExistForVlan, BadVrrpTimers, BadVrrpTracking, UnknownDhcpRelayServer, DhcpRelayServerAlreadyExists, \
    VlanAlreadyExist, UnknownBond, InvalidAccessGroupName, InvalidUnicastRPFMode, UnsupportedOperation
from netman.core.objects.interface import Interface
from netman.core.objects.interface_states import OFF, ON
from netman.core.objects.port_modes import DYNAMIC, ACCESS, TRUNK
from netman.core.objects.switch_base import SwitchBase
from netman.core.objects.switch_transactional import FlowControlSwitch
from netman.core.objects.unicast_rpf_modes import STRICT
from netman.core.objects.vlan import Vlan
from netman.core.objects.vrrp_group import VrrpGroup


def ssh(switch_descriptor):
    return Cisco(switch_descriptor=switch_descriptor)


def factory(switch_descriptor, lock):
    warnings.warn("Use SwitchFactory.get_switch_by_descriptor directly to instantiate a switch", DeprecationWarning)

    return FlowControlSwitch(wrapped_switch=ssh(switch_descriptor), lock=lock)


class Cisco(SwitchBase):

    def __init__(self, switch_descriptor):
        super(Cisco, self).__init__(switch_descriptor)
        self.ssh = None

    def _connect(self):
        params = dict(
            host=self.switch_descriptor.hostname,
            username=self.switch_descriptor.username,
            password=self.switch_descriptor.password,
        )
        if self.switch_descriptor.port:
            params["port"] = self.switch_descriptor.port

        self.ssh = SshClient(**params)

        if self.ssh.get_current_prompt().endswith(">"):
            self.ssh.do("enable", wait_for=": ")
            self.ssh.do(self.switch_descriptor.password)

        self.ssh.do("terminal length 0")
        self.ssh.do("terminal width 0")

    def _disconnect(self):
        self.ssh.quit("exit")
        self.logger.info(self.ssh.full_log)

    def _end_transaction(self):
        pass

    def _start_transaction(self):
        pass

    def commit_transaction(self):
        self.ssh.do("write memory")

    def rollback_transaction(self):
        pass

    def get_vlan(self, number):
        vlan = Vlan(int(number), icmp_redirects=True, arp_routing=True, ntp=True)
        apply_vlan_running_config_data(vlan, self._get_vlan_run_conf(number))
        apply_interface_running_config_data(
            vlan,
            self.ssh.do("show running-config interface vlan {} | begin interface".format(number))
        )
        return vlan

    def get_vlans(self):
        vlan_list = self.ssh.do("show vlan brief")

        vlans = {}
        for line in vlan_list:
            if regex.match('^(\d+)\s+(\S+).*', line):
                number, name = regex

                if name == ("VLAN{}".format(number)):
                    name = None

                vlans[number] = Vlan(int(number), name, icmp_redirects=True, arp_routing=True, ntp=True)

        for ip_interface_data in split_on_dedent(self.ssh.do("show ip interface")):
            if regex.match("^Vlan(\d+)\s.*", ip_interface_data[0]):
                current_vlan = vlans.get(regex[0])
                if current_vlan:
                    apply_interface_running_config_data(
                        current_vlan,
                        self.ssh.do("show running-config interface vlan {}".format(current_vlan.number))
                    )
        return vlans.values()

    def add_vlan(self, number, name=None):
        if self._show_run_vlan(number):
            raise VlanAlreadyExist(number)

        with self.config():
            result = self.ssh.do('vlan {}'.format(number))
            if len(result) > 0:
                raise BadVlanNumber()
            else:
                if name:
                    result = self.ssh.do('name {}'.format(name))

                self.ssh.do('exit')

                if len(result) > 0:
                    raise BadVlanName()

    def remove_vlan(self, number):
        self._get_vlan_run_conf(number)

        with self.config():
            self.ssh.do('no interface vlan {}'.format(number))
            self.ssh.do('no vlan {}'.format(number))

    def get_interfaces(self):
        interfaces = []
        for data in split_on_bang(self.ssh.do("show running-config | begin interface")):
            interface = parse_interface(data)
            if interface:
                interfaces.append(interface)
        return interfaces

    def set_access_vlan(self, interface_id, vlan):
        self._get_vlan_run_conf(vlan)

        with self.config(), self.interface(interface_id):
            self.ssh.do('switchport access vlan {}'.format(vlan))

    def unset_interface_access_vlan(self, interface_id):
        with self.config(), self.interface(interface_id):
            self.ssh.do('no switchport access vlan')

    def set_access_mode(self, interface_id):
        with self.config(), self.interface(interface_id):
            self.ssh.do('switchport mode access')
            self.ssh.do('no switchport trunk native vlan')
            self.ssh.do('no switchport trunk allowed vlan')

    def set_trunk_mode(self, interface_id):
        has_trunk_allowed = self.has_trunk_allowed_set(interface_id)
        with self.config(), self.interface(interface_id):
            self.ssh.do('switchport mode trunk')
            if not has_trunk_allowed:
                self.ssh.do('switchport trunk allowed vlan none')
            self.ssh.do('no switchport access vlan')

    def add_trunk_vlan(self, interface_id, vlan):
        self._get_vlan_run_conf(vlan)

        with self.config(), self.interface(interface_id):
            self.ssh.do('switchport trunk allowed vlan add {}'.format(vlan))

    def remove_trunk_vlan(self, interface_id, vlan):
        interface = self.get_interface(interface_id)
        if vlan not in interface.trunk_vlans:
            raise UnknownVlan(vlan)

        with self.config(), self.interface(interface_id):
            self.ssh.do('switchport trunk allowed vlan remove {}'.format(vlan))

    def set_interface_state(self, interface_id, state):
        with self.config(), self.interface(interface_id):
            self.ssh.do('shutdown' if state is OFF else "no shutdown")

    def set_interface_native_vlan(self, interface_id, vlan):
        self._get_vlan_run_conf(vlan)

        with self.config(), self.interface(interface_id):
            self.ssh.do('switchport trunk native vlan {}'.format(vlan))

    def unset_interface_native_vlan(self, interface_id):
        with self.config(), self.interface(interface_id):
            self.ssh.do('no switchport trunk native vlan')

    def add_ip_to_vlan(self, vlan_number, ip_network):
        vlan = self.get_vlan_interface_data(vlan_number)

        ip_found = next((ip for ip in vlan.ips if ip.ip == ip_network.ip), False)
        if ip_found:
            raise IPAlreadySet(ip_network, ip_found)

        has_ips = len(vlan.ips) > 0
        with self.config(), self.interface_vlan(vlan_number):
            if has_ips:
                self.ssh.do('no ip redirects')
            result = self.ssh.do('ip address {} {}{}'.format(ip_network.ip, ip_network.netmask,
                                                             " secondary" if has_ips else ""))
            if len(result) > 0:
                raise IPNotAvailable(ip_network, reason="; ".join(result))

    def remove_ip_from_vlan(self, vlan_number, ip_network):
        vlan = self.get_vlan_interface_data(vlan_number)
        existing_ip = next((ip for ip in vlan.ips if ip.ip == ip_network.ip and ip.netmask == ip_network.netmask), False)

        if existing_ip:
            ip_index = vlan.ips.index(existing_ip)
            with self.config(), self.interface_vlan(vlan_number):
                if ip_index == 0:
                    if len(vlan.ips) == 1:
                        self.ssh.do('no ip address {} {}'.format(ip_network.ip, ip_network.netmask))
                    else:
                        self.ssh.do('ip address {} {}'.format(vlan.ips[1].ip, vlan.ips[1].netmask))
                else:
                    self.ssh.do('no ip address {} {} secondary'.format(ip_network.ip, ip_network.netmask))
        else:
            raise UnknownIP(ip_network)

    def has_trunk_allowed_set(self, interface_id):
        for line in self.ssh.do('show running-config interface {}'.format(interface_id)):
            if 'Invalid input detected' in line:
                raise UnknownInterface(interface_id)
            if 'switchport trunk allowed' in line:
                return True
        return False

    def set_vlan_access_group(self, vlan_number, direction, name):
        self.get_vlan_interface_data(vlan_number)

        with self.config(), self.interface_vlan(vlan_number):
            result = self.ssh.do("ip access-group {} {}".format(name, 'in' if direction == IN else 'out'))
            if len(result) > 0:
                raise InvalidAccessGroupName(name)

    def unset_vlan_access_group(self, vlan_number, direction):
        vlan = self.get_vlan_interface_data(vlan_number)

        if vlan.access_groups[direction] is None:
            raise UnknownAccessGroup(direction)
        else:
            with self.config(), self.interface_vlan(vlan_number):
                self.ssh.do("no ip access-group {}".format(('in' if direction == IN else 'out')))

    def set_vlan_vrf(self, vlan_number, vrf_name):
        self.get_vlan_interface_data(vlan_number)

        with self.config(), self.interface_vlan(vlan_number):
            result = self.ssh.do("ip vrf forwarding {}".format(vrf_name))
            if len(result) > 0:
                raise UnknownVrf(vrf_name)

    def unset_vlan_vrf(self, vlan_number):
        vlan = self.get_vlan_interface_data(vlan_number)

        if vlan.vrf_forwarding is None:
            raise VlanVrfNotSet(vlan_number)
        else:
            with self.config(), self.interface_vlan(vlan_number):
                self.ssh.do("no ip vrf forwarding")

    def add_dhcp_relay_server(self, vlan_number, ip_address):
        vlan = self.get_vlan_interface_data(vlan_number)

        if ip_address in vlan.dhcp_relay_servers:
            raise DhcpRelayServerAlreadyExists(vlan_number=vlan_number, ip_address=ip_address)

        with self.config(), self.interface_vlan(vlan_number):
            self.ssh.do("ip helper-address {}".format(ip_address))

    def remove_dhcp_relay_server(self, vlan_number, ip_address):
        vlan = self.get_vlan_interface_data(vlan_number)

        if ip_address not in vlan.dhcp_relay_servers:
            raise UnknownDhcpRelayServer(vlan_number=vlan_number, ip_address=ip_address)

        with self.config(), self.interface_vlan(vlan_number):
            self.ssh.do("no ip helper-address {}".format(ip_address))

    def get_vlan_interfaces(self, vlan_number):
        vlan_interfaces = get_vlan_interfaces_from_data(vlan_number, self.get_interfaces())
        if not vlan_interfaces:
            self.get_vlan(vlan_number)
        return vlan_interfaces

    def set_bond_trunk_mode(self, number):
        with NamedBond(number) as bond:
            return self.set_trunk_mode(bond.name)

    def set_bond_access_mode(self, number):
        with NamedBond(number) as bond:
            return self.set_access_mode(bond.name)

    def add_bond_trunk_vlan(self, number, vlan):
        with NamedBond(number) as bond:
            return self.add_trunk_vlan(bond.name, vlan)

    def remove_bond_trunk_vlan(self, number, vlan):
        with NamedBond(number) as bond:
            return self.remove_trunk_vlan(bond.name, vlan)

    def set_bond_native_vlan(self, number, vlan):
        with NamedBond(number) as bond:
            return self.set_interface_native_vlan(bond.name, vlan)

    def unset_bond_native_vlan(self, number):
        with NamedBond(number) as bond:
            return self.unset_interface_native_vlan(bond.name)

    def config(self):
        return SubShell(self.ssh, enter="configure terminal", exit_cmd='exit')

    def interface(self, interface_id):
        return SubShell(self.ssh, enter="interface {}".format(interface_id), exit_cmd='exit',
                        validate=no_output(UnknownInterface, interface_id))

    def interface_vlan(self, interface_id):
        return SubShell(self.ssh, enter=["interface vlan {}".format(interface_id), "no shutdown"], exit_cmd='exit')

    def _get_vlan_run_conf(self, vlan_number):
        run_config = self._show_run_vlan(vlan_number)
        if not run_config:
            raise UnknownVlan(vlan_number)
        return run_config

    def _show_run_vlan(self, vlan_number):
        return self.ssh.do('show running-config vlan {} | begin vlan'.format(vlan_number))

    def get_vlan_interface_data(self, vlan_number):
        run_int_vlan_data = self.ssh.do('show running-config interface vlan {}'.format(vlan_number))
        if not run_int_vlan_data[0].startswith("Building configuration..."):
            self._get_vlan_run_conf(vlan_number)

        vlan = Vlan(vlan_number)
        apply_interface_running_config_data(vlan, run_int_vlan_data)
        return vlan

    def get_interface(self, interface_id):
        data = self.ssh.do("show running-config interface {} | begin interface".format(interface_id))
        interface = parse_interface(data)
        if not interface:
            raise UnknownInterface(interface_id)
        return interface

    def add_vrrp_group(self, vlan_number, group_id, ips=None, priority=None, hello_interval=None, dead_interval=None,
                       track_id=None, track_decrement=None):
        if not (0 < group_id <= 255):
            raise BadVrrpGroupNumber(1, 255)

        vlan = self.get_vlan_interface_data(vlan_number)

        if [group for group in vlan.vrrp_groups if group.id == group_id]:
            raise VrrpAlreadyExistsForVlan(vlan=vlan_number, vrrp_group_id=group_id)

        with self.config(), self.interface_vlan(vlan_number):
            if len(vlan.vrrp_groups) == 0:
                self.ssh.do('standby version 2')

            if hello_interval is not None and dead_interval is not None:
                result = self.ssh.do('standby {group_id} timers {hello_interval} {dead_interval}'.format(
                    group_id=group_id, hello_interval=hello_interval, dead_interval=dead_interval))
                if len(result) > 0:
                    raise BadVrrpTimers()

            if priority is not None:
                result = self.ssh.do('standby {group_id} priority {priority}'.format(group_id=group_id, priority=priority))
                if len(result) > 0:
                    raise BadVrrpPriorityNumber(1, 255)

            self.ssh.do('standby {group_id} preempt delay minimum 60'.format(group_id=group_id))
            self.ssh.do('standby {group_id} authentication {authentication}'.format(group_id=group_id,
                                                                                    authentication='VLAN{}'.format(vlan_number)))

            if track_id is not None and track_decrement is not None:
                result = self.ssh.do('standby {group_id} track {track_id} decrement {track_decrement}'.format(
                    group_id=group_id, track_id=track_id, track_decrement=track_decrement))
                if len(result) > 0:
                    raise BadVrrpTracking()

            for i, ip in enumerate(ips):
                result = self.ssh.do('standby {group_id} ip {ip}{secondary}'.format(
                    group_id=group_id, ip=ip, secondary=' secondary' if i > 0 else ''))
                if len(result) > 0:
                    raise IPNotAvailable(ip, reason="; ".join(result))

    def remove_vrrp_group(self, vlan_number, group_id):
        vlan = self.get_vlan_interface_data(vlan_number)

        if not [group for group in vlan.vrrp_groups if group.id == group_id]:
            raise VrrpDoesNotExistForVlan(vlan=vlan_number, vrrp_group_id=group_id)

        with self.config(), self.interface_vlan(vlan_number):
            result = self.ssh.do('no standby {group_id}'.format(group_id=group_id))
            if len(result) > 0:
                raise VrrpDoesNotExistForVlan(vlan=vlan_number, vrrp_group_id=group_id)

            if len(vlan.vrrp_groups) == 1:
                self.ssh.do('no standby version')

    def set_vlan_arp_routing_state(self, vlan_number, state):
        self.get_vlan_interface_data(vlan_number)

        with self.config(), self.interface_vlan(vlan_number):
            if state == ON:
                self.ssh.do('ip proxy-arp')
            else:
                self.ssh.do('no ip proxy-arp')

    def set_vlan_icmp_redirects_state(self, vlan_number, state):
        self.get_vlan_interface_data(vlan_number)

        with self.config(), self.interface_vlan(vlan_number):
            if state:
                self.ssh.do('ip redirects')
            else:
                self.ssh.do('no ip redirects')

    def set_vlan_ntp_state(self, vlan_number, state):
        self.get_vlan_interface_data(vlan_number)

        with self.config(), self.interface_vlan(vlan_number):
            self.ssh.do('{}ntp disable'.format('no ' if state else ''))

    def reset_interface(self, interface_id):
        with self.config():
            for line in self.ssh.do('default interface {}'.format(interface_id)):
                if 'Invalid input detected' in line:
                    raise UnknownInterface(interface_id)

    def get_versions(self):
        result = self.ssh.do('show version')

        versions = {}
        for i, line in enumerate(result):
            matches = re.match("^(.*)\s:\s(.*)$", line)
            if matches:
                values = matches.groups()
                versions[values[0].strip()] = values[1]

            matches = re.match("^.*(\d+)\s+(\d+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s*$", line)
            if matches:
                values = matches.groups()
                if "units" not in versions:
                    versions["units"] = {}

                versions["units"][values[0]] = {
                    "Ports": values[1],
                    "Model": values[2],
                    "SW Version": values[3],
                    "SW Image": values[4]
                }

        return versions

    def set_vlan_unicast_rpf_mode(self, vlan_number, mode):
        operations = {STRICT: self._set_unicast_rpf_strict}

        if mode not in operations:
            raise InvalidUnicastRPFMode(mode)

        self.get_vlan_interface_data(vlan_number)

        with self.config(), self.interface_vlan(vlan_number):
            operations[mode]()

    def unset_vlan_unicast_rpf_mode(self, vlan_number):
        self.get_vlan_interface_data(vlan_number)

        with self.config(), self.interface_vlan(vlan_number):
            self.ssh.do('no ip verify unicast')

    def _set_unicast_rpf_strict(self):
        result = self.ssh.do('ip verify unicast source reachable-via rx')
        if len(result) > 0:
            raise UnsupportedOperation("Unicast RPF Mode Strict", "\n".join(result))


def parse_interface(data):
    if data and (regex.match("interface (\w*Ethernet[^\s]*)", data[0]) or regex.match("interface (Port-channel[^\s]*)", data[0])):
        i = Interface(name=regex[0], shutdown=False)
        port_mode = access_vlan = native_vlan = trunk_vlans = None
        for line in data:
            if regex.match(" switchport mode (.*)", line):
                port_mode = regex[0]
            if regex.match(" switchport access vlan (\d*)", line):
                access_vlan = int(regex[0])
            if regex.match(" switchport trunk native vlan (\d*)", line):
                native_vlan = int(regex[0])
            if regex.match(" switchport trunk allowed vlan (.*)", line):
                trunk_vlans = regex[0]
            if regex.match(" shutdown", line):
                i.shutdown = True

        if not port_mode:
            i.port_mode = DYNAMIC
            i.access_vlan = access_vlan
            i.trunk_native_vlan = native_vlan
            i.trunk_vlans = parse_vlan_ranges(trunk_vlans) if trunk_vlans else []
        elif port_mode == 'access':
            i.port_mode = ACCESS
            i.access_vlan = access_vlan
        elif port_mode == 'trunk':
            i.port_mode = TRUNK
            i.trunk_native_vlan = native_vlan
            i.trunk_vlans = parse_vlan_ranges(trunk_vlans) if trunk_vlans else []

        return i
    return None


def apply_interface_running_config_data(vlan, data):
    for line in data:
        if regex.match("^ ip address ([^\s]*) ([^\s]*)(.*)", line):
            ip = IPNetwork("{}/{}".format(regex[0], regex[1]))
            if "secondary" not in regex[2]:
                vlan.ips.insert(0, ip)
            else:
                vlan.ips.append(ip)

        elif regex.match("^ ip access-group ([^\s]*) ([^\s]*).*", line):
            if regex[1] == "in":
                vlan.access_groups[IN] = regex[0]
            else:
                vlan.access_groups[OUT] = regex[0]

        elif regex.match("^ ip vrf forwarding ([^\s]*).*", line):
            vlan.vrf_forwarding = regex[0]

        elif regex.match("^ standby ([\d]+) (.*)", line):
            vrrp_group = next((group for group in vlan.vrrp_groups if str(group.id) == regex[0]), None)
            if vrrp_group is None:
                vrrp_group = VrrpGroup(id=int(regex[0]))
                vlan.vrrp_groups.append(vrrp_group)

            vrrp_info = regex[1].strip()

            if regex.match("^ip ([^\s]*).*", vrrp_info):
                vrrp_group.ips.append(IPAddress(regex[0]))
            elif regex.match("^timers ([^\s]*) ([^\s]*)", vrrp_info):
                vrrp_group.hello_interval = int(regex[0])
                vrrp_group.dead_interval = int(regex[1])
            elif regex.match("^priority ([^\s]*)", vrrp_info):
                vrrp_group.priority = int(regex[0])
            elif regex.match("^track ([^\s]*) decrement ([^\s]*)", vrrp_info):
                vrrp_group.track_id = regex[0]
                vrrp_group.track_decrement = int(regex[1])

        elif regex.match("^ ip helper-address ([^\s]*)", line):
            vlan.dhcp_relay_servers.append(IPAddress(regex[0]))

        elif regex.match("^ no ip proxy-arp", line):
            vlan.arp_routing = False

        elif regex.match("^ no ip redirects", line):
            vlan.icmp_redirects = False

        elif regex.match("^ ip verify unicast source reachable-via rx", line):
            vlan.unicast_rpf_mode = STRICT

        elif regex.match("^ ntp disable", line):
            vlan.ntp = False


def apply_vlan_running_config_data(vlan, data):
    for line in data:
        if regex.match("^ name ([^\s]*)", line):
            vlan.name = regex[0]


def parse_vlan_ranges(all_ranges):
    if all_ranges is None:
        return range(1, 4094)
    elif all_ranges == "none":
        return []
    else:
        full_list = []
        for vlan_list in [parse(r) for r in all_ranges.split(",")]:
            full_list += vlan_list
        return full_list


def get_vlan_interfaces_from_data(vlan_number, interfaces_data):
    vlan_interfaces = []
    for interface in interfaces_data:
        if interface.port_mode is TRUNK:
            if interface.trunk_vlans and vlan_number in interface.trunk_vlans:
                vlan_interfaces.append(interface.name)
            elif vlan_number == interface.trunk_native_vlan:
                vlan_interfaces.append(interface.name)
        elif interface.port_mode is ACCESS and interface.access_vlan == vlan_number:
            vlan_interfaces.append(interface.name)
        elif interface.port_mode is DYNAMIC:
            if interface.access_vlan == vlan_number or interface.trunk_native_vlan == vlan_number or \
                    (interface.trunk_vlans and vlan_number in interface.trunk_vlans):
                vlan_interfaces.append(interface.name)
    return vlan_interfaces


def parse(single_range):
    if regex.match("(\d+)-(\d+)", single_range):
        return range(int(regex[0]), int(regex[1]) + 1)
    else:
        return [int(single_range)]


def bond_name(number):
    return "Port-channel{}".format(number)


class NamedBond(object):
    def __init__(self, number):
        self.number = number

    @property
    def name(self):
        return bond_name(self.number)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is UnknownInterface:
            raise UnknownBond(self.number), None, exc_tb



# Copyright 2019 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hamcrest import assert_that, is_, empty
from netaddr import IPNetwork

from netman.core.objects.exceptions import UnknownIP, BadVlanNumber
from tests.adapters.compliance_test_case import ComplianceTestCase


class RemoveIpFromVlanTest(ComplianceTestCase):
    _dev_sample = "arista_http"

    def setUp(self):
        super(RemoveIpFromVlanTest, self).setUp()
        self.try_to.add_vlan(1000, name="some_name")
        self.try_to.add_ip_to_vlan(1000, IPNetwork("10.10.10.2/29"))

    def tearDown(self):
        self.janitor.remove_ip_from_vlan(1000, IPNetwork("10.10.10.2/29"))
        self.janitor.remove_vlan(1000)

        super(RemoveIpFromVlanTest, self).tearDown()

    def test_unassigns(self):
        self.client.remove_ip_from_vlan(1000, IPNetwork("10.10.10.2/29"))
        assert_that(self.client.get_vlan(1000).ips, is_(empty()))

    def test_raise_bad_vlan_number(self):
        with self.assertRaises(BadVlanNumber):
            self.client.remove_ip_from_vlan(0, IPNetwork("10.10.10.2/29"))

    def test_raise_unknown_ip(self):
        with self.assertRaises(UnknownIP):
            self.client.remove_ip_from_vlan(1000, IPNetwork("1.1.1.1/31"))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from hamcrest import assert_that, is_

from netman.core.objects.exceptions import UnknownVlan
from netman.core.objects.unicast_rpf_modes import STRICT
from tests import has_message
from tests.adapters.compliance_test_case import ComplianceTestCase


class SetVlanUnicastRPFModeTest(ComplianceTestCase):
    _dev_sample = "cisco"

    def setUp(self):
        super(SetVlanUnicastRPFModeTest, self).setUp()
        self.client.add_vlan(1000)

    def test_can_activate_the_strict_unicast_anti_spoofing(self):
        self.client.set_vlan_unicast_rpf_mode(1000, STRICT)

        vlan = self.get_vlan_from_list(1000)

        assert_that(vlan.unicast_rpf_mode, is_(STRICT))

    def test_raises_UnknownVlan_when_operating_on_a_vlan_that_does_not_exist(self):
        with self.assertRaises(UnknownVlan) as expect:
            self.client.set_vlan_unicast_rpf_mode(2000, STRICT)

        assert_that(expect.exception, has_message("Vlan 2000 not found"))

    def tearDown(self):
        self.janitor.remove_vlan(1000)
        super(SetVlanUnicastRPFModeTest, self).tearDown()



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from flask import send_from_directory
from pkg_resources import get_distribution

from netman.api.api_utils import to_response
from netman.api.objects import info


class NetmanApi(object):
    def __init__(self, switch_factory=None, get_distribution_callback=get_distribution):
        self.switch_factory = switch_factory
        self.app = None
        self.get_distribution = get_distribution_callback

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def hook_to(self, server):
        self.app = server
        server.add_url_rule('/netman/info', endpoint="netman_info", view_func=self.get_info, methods=['GET'])
        server.add_url_rule('/netman/apidocs/', endpoint="netman_apidocs", view_func=self.api_docs, methods=['GET'])
        server.add_url_rule('/netman/apidocs/<path:filename>', endpoint="netman_apidocs", view_func=self.api_docs, methods=['GET'])

    @to_response
    def get_info(self):
        """
        Informations about the current deployment and state and generates a log entry on the netman.api logger \
        that says : ``/info requested this is a logging test``

        :code 200 OK:

        Example output:

        .. literalinclude:: ../doc_config/api_samples/get_info.json
            :language: json

        """
        logging.getLogger("netman.api").info("/info requested this is a logging test")

        return 200, info.to_api(
            status='running',
            version=self.get_distribution('netman').version,
            lock_provider=_class_fqdn(self.switch_factory.lock_factory)
        )

    def api_docs(self, filename=None):
        """
        Shows this documentation

        """
        return send_from_directory(os.path.dirname(__file__) + "/doc/html/", filename or "index.html")


def _class_fqdn(obj):
    return "{}.{}".format(obj.__module__, obj.__class__.__name__)



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hamcrest import assert_that, is_

from netman.core.objects.exceptions import UnknownInterface
from netman.core.objects.interface_states import ON
from tests import has_message
from tests.adapters.compliance_test_case import ComplianceTestCase


class GetInterfaceTest(ComplianceTestCase):
    _dev_sample = "juniper_mx"

    def setUp(self):
        super(GetInterfaceTest, self).setUp()

    def test_returns_an_interface(self):
        interface = self.client.get_interface(self.test_ports[0].name)
        assert_that(interface.name, is_(self.test_ports[0].name))

    def test_get_interface_and_get_interfaces_are_same(self):
        self.client.add_vlan(1000, name="vlan1000")
        self.client.add_vlan(2000, name="vlan2000")
        expected = self.test_ports[0]

        self.try_to.set_access_vlan(expected.name, 1000)
        self.try_to.set_trunk_mode(expected.name)
        self.try_to.set_interface_state(expected.name, ON)
        self.try_to.set_interface_native_vlan(expected.name, 2000)
        self.try_to.set_interface_auto_negotiation_state(expected.name, ON)
        self.try_to.set_interface_mtu(expected.name, 5000)

        interface_from_single = self.client.get_interface(expected.name)
        interfaces = [inte for inte in self.client.get_interfaces() if inte.name == expected.name]
        interface_from_multiple = interfaces[0]

        assert_that(interface_from_single.name, is_(interface_from_multiple.name))
        assert_that(interface_from_single.port_mode, is_(interface_from_multiple.port_mode))
        assert_that(interface_from_single.shutdown, is_(interface_from_multiple.shutdown))
        assert_that(interface_from_single.trunk_native_vlan, is_(interface_from_multiple.trunk_native_vlan))
        assert_that(interface_from_single.trunk_vlans, is_(interface_from_multiple.trunk_vlans))
        assert_that(interface_from_single.auto_negotiation, is_(interface_from_multiple.auto_negotiation))
        assert_that(interface_from_single.mtu, is_(interface_from_multiple.mtu))

    def test_getinterface_nonexistent_raises(self):
        with self.assertRaises(UnknownInterface)as expect:
            self.client.get_interface('ethernet 1/nonexistent2000')

        assert_that(expect.exception, has_message("Unknown interface ethernet 1/nonexistent2000"))

    def tearDown(self):
        self.janitor.unset_interface_access_vlan(self.test_ports[0].name)
        self.janitor.unset_interface_native_vlan(self.test_ports[0].name)
        self.janitor.set_access_mode(self.test_ports[0].name)

        self.janitor.remove_vlan(1000)
        self.janitor.remove_vlan(2000)
        super(GetInterfaceTest, self).tearDown()



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hamcrest import assert_that, is_
from netman.core.objects.exceptions import UnknownInterface
from netman.core.objects.interface_states import ON, OFF
from tests import has_message
from tests.adapters.compliance_test_case import ComplianceTestCase


class UnsetInterfaceStateTest(ComplianceTestCase):
    _dev_sample = "juniper"

    def test_returns_an_interface_to_default_state(self):

        default_state = self.client.get_interface(self.test_port).shutdown

        self.try_to.set_interface_state(self.test_port, ON)
        self.client.unset_interface_state(self.test_port)

        assert_that(self.client.get_interface(self.test_port).shutdown, is_(default_state))

        self.try_to.set_interface_state(self.test_port, OFF)
        self.client.unset_interface_state(self.test_port)

        assert_that(self.client.get_interface(self.test_port).shutdown, is_(default_state))

    def test_unset_an_interface_twice_works(self):
        default_state = self.client.get_interface(self.test_port).shutdown
        self.try_to.set_interface_state(self.test_port, ON)

        self.client.unset_interface_state(self.test_port)
        self.client.unset_interface_state(self.test_port)

        self.try_to.set_interface_state(self.test_port, default_state)
        assert_that(self.client.get_interface(self.test_port).shutdown, is_(default_state))

    def test_fails_with_unknown_interface(self):
        with self.assertRaises(UnknownInterface) as expect:
            self.client.unset_interface_state('ge-0/0/1nonexistent2000')

        assert_that(expect.exception, has_message("Unknown interface ge-0/0/1nonexistent2000"))

    def tearDown(self):
        super(UnsetInterfaceStateTest, self).tearDown()



from hamcrest import assert_that, is_

from netman.core.objects.exceptions import UnknownInterface
from netman.core.objects.interface_states import ON, OFF
from netman.core.objects.port_modes import ACCESS
from tests.adapters.compliance_test_case import ComplianceTestCase


class ResetInterfaceTest(ComplianceTestCase):
    _dev_sample = "juniper"

    def setUp(self):
        super(ResetInterfaceTest, self).setUp()
        self.interface_before = self.client.get_interface(self.test_port)

    def tearDown(self):
        self.janitor.remove_vlan(90)
        self.janitor.remove_bond(42)
        super(ResetInterfaceTest, self).tearDown()

    def test_reverts_trunk_vlans(self):
        self.try_to.add_vlan(90)

        self.try_to.set_trunk_mode(self.test_port)
        self.try_to.add_trunk_vlan(self.test_port, 90)

        self.client.reset_interface(self.test_port)

        assert_that(self.client.get_interface(self.test_port).trunk_vlans, is_(self.interface_before.trunk_vlans))

    def test_reverts_trunk_native_vlan(self):
        self.try_to.add_vlan(90)

        self.try_to.set_trunk_mode(self.test_port)
        self.try_to.set_interface_native_vlan(self.test_port, 90)

        self.client.reset_interface(self.test_port)

        assert_that(self.client.get_interface(self.test_port).trunk_native_vlan,
                    is_(self.interface_before.trunk_native_vlan))

    def test_reverts_port_mode(self):
        if self.interface_before.port_mode == ACCESS:
            self.try_to.set_trunk_mode(self.test_port)
        else:
            self.try_to.set_access_mode(self.test_port)

        self.client.reset_interface(self.test_port)

        assert_that(self.try_to.get_interface(self.test_port).port_mode, is_(self.interface_before.port_mode))

    def test_reverts_bond_master(self):
        self.try_to.add_bond(42)

        self.try_to.add_interface_to_bond(self.test_port, 42)

        self.client.reset_interface(self.test_port)

        actual_interface = self.try_to.get_interface(self.test_port)
        assert_that(actual_interface.bond_master, is_(self.interface_before.bond_master))

    def test_reverts_port_shutdown(self):
        if not self.interface_before.shutdown or self.interface_before.shutdown is OFF:
            self.try_to.set_interface_state(self.test_port, ON)
        else:
            self.try_to.set_interface_state(self.test_port, OFF)

        self.client.reset_interface(self.test_port)

        assert_that(self.client.get_interface(self.test_port).shutdown, is_(self.interface_before.shutdown))

    def test_reset_interface_mtu(self):
        self.try_to.set_interface_mtu(self.test_port, 4000)

        self.client.reset_interface(self.test_port)

        assert_that(self.client.get_interface(self.test_port).mtu, is_(self.interface_before.mtu))

    def test_raises_on_unknown_interface(self):
        with self.assertRaises(UnknownInterface):
            self.client.reset_interface('nonexistent 2/99')



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netman.core.objects import Model


class SwitchDescriptor(Model):
    def __init__(self, model, hostname, username=None, password=None, port=None, netman_server=None):
        self.model = model
        self.hostname = hostname
        self.username = username
        self.password = password
        self.port = port
        self.netman_server = netman_server
